#! /bin/bash
# mpath_connmark.sh: configure settings for ISP links
# Usage: $0 config_file [{apply|stop}]
# Assume "showcfg" by default

cmd="$2"
cfg="$1"
if [ "$cfg" = "" ]
then
	echo "Usage: $0 config_file [{apply|stop}]"
	exit 1
fi

if ! [ -e "$cfg" ]
then
	echo "Missing configuration file: $cfg"
	exit 1
fi

dp=""
cd "`dirname "$0"`" && dp="`pwd`"

. "$cfg"

if [ "$dp" = "" ]
then
	echo "Unknown script dir"
	exit 1
fi

test=y
[ "$cmd" != "" ] && test=""

if [ "$test" = y ]
then
	IP="echo ip"
	IPA="echo ip"
	IPT="echo iptables"
	IPTA="echo iptables"
	TEST="echo"
else
	# ip
	IP="ip"
	# ip ... add
	IPA="ip"
	[ "$cmd" = stop ] && IPA="true"
	# iptables
	IPT="iptables"
	# iptables ... -A
	IPTA="iptables"
	[ "$cmd" = stop ] && IPTA="true"
	TEST=""
fi

function log() {
	echo "# $@"
}

exit_func() {
	$IPT -t mangle -D PREROUTING -j RELOAD
	$IPT -t mangle -D OUTPUT -j RELOAD
	$IPT -t mangle -F RELOAD
	$IPT -t mangle -X RELOAD
	rm -f "$LOCK"
}

# Back in time?
find $LOCK -mmin -0 -exec rm {} \; &> /dev/null
# Lock rules
lockfile -l 120 -s 1 -r 0 -1 $LOCK &> /dev/null || exit 1

trap exit_func EXIT

log "Stop traffic while reloading:"
$IPT -t mangle -N RELOAD
$IPT -t mangle -A RELOAD -j DROP
$IPT -t mangle -I PREROUTING -j RELOAD
$IPT -t mangle -I OUTPUT -j RELOAD

log "Remove old ip rules 1..100:"
while IFS=":" read pref rule_part
do
	[ "$pref" -ge 1 -a "$pref" -le 100 ] && \
		$IP rule del pref $pref
done < <(ip rule)

log "Remove old routes:"
for ((lnk=0; lnk < ${#LINKS[@]}; lnk++))
do
	$IP route del ${GW[$lnk]}/32 dev ${EIF[$lnk]}
	#$IP route flush dev ${EIF[$lnk]} table main scope global
	$IP route flush table $[$lnk+1]
done
$IP route flush table 100

### Routing rules (in creation order):

log "Routes to direct gateways in table main:"
for ((lnk=0; lnk < ${#LINKS[@]}; lnk++))
do
	[ "${LINKS[$lnk]}" != y ] && continue
	$IPA route add ${GW[$lnk]}/32 dev ${EIF[$lnk]} src ${LIP[$lnk]}
done

log "Try routes to local subnets in table main:"
$IPA rule add pref 5 to 0/0 lookup main

# 
log "Routes via specific links, usually for established conns or known dests:"
for ((lnk=0; lnk < ${#LINKS[@]}; lnk++))
do
	if [ "${LINKS[$lnk]}" = y ]
	then
		$IPA route add default via ${GW[$lnk]} dev ${EIF[$lnk]} \
			src ${PIP[$lnk]} table $[$lnk+1]
	fi
	# By default, drop traffic if device is unregistered
	$IPA route add prohibit default metric 1 table $[$lnk+1]
done

log "Route outgoing traffic for established connections via their assigned link:"
for ((lnk=0; lnk < ${#LINKS[@]}; lnk++))
do
	[ "${LINKS[$lnk]}" != y ] && continue
	$IPA rule add pref $[11+$lnk] from 0/0 fwmark $[$lnk+1] lookup $[$lnk+1]
done

log "New connections for known destinations via specific link, eg. indirect GWs:"
for ((lnk=0; lnk < ${#LINKS[@]}; lnk++))
do
	[ "${LINKS[$lnk]}" != y ] && continue
	[ "${CHKGW[$lnk]}" = "" ] && continue
	$IPA rule add pref $[21+$lnk] to ${CHKGW[$lnk]} lookup $[$lnk+1]
done

log "Public IPs prefer specific link for traffic to world:"
for ((lnk=0; lnk < ${#LINKS[@]}; lnk++))
do
	[ "${LINKS[$lnk]}" != y ] && continue
	$IPA rule add pref $[31+$lnk] from ${PIP[$lnk]} lookup $[$lnk+1]
done

# Default routes, usually multipath, used by private sources or
# source address autoselection (source 0.0.0.0).
# Actualized by scripts depending on health checks.
# All routes in table 100 should be recreated by scripts if some used device
# is unregistered (device disappears) because the kernel removes such routes.

log "Reset default routes in table 100:"
$IPA route add prohibit default metric 1 table 100 2> /dev/null
$IPA rule add pref 100 from 0/0 lookup 100

log "Remove default route from table main:"
$IPA route del default table main

log "Flush route cache:"
$IP route flush cache

# Create tables for enabled links: LINK1, LINK2, etc
# They can be used for custom iptables rules that may use specific
# links in some order. Example for port 25 is below.
# Start with no rules, mpath_connmark_check.sh and
# mpath_connmark_change_links.sh should maintain rules in these tables.
for ((lnk=0; lnk < ${#LINKS[@]}; lnk++))
do
	[ "${LINKS[$lnk]}" != y ] && continue
	$IPTA -t mangle -N LINK$[$lnk+1]
done

# PREROUTING mangle:

## INPUT_SELECTION table (mangle):
# - called for PREROUTING (before DNAT) when mark is 0, for
# new incoming connections:
#	- select outgoing link or leave it to routing
$IPT -t mangle -F INPUT_SELECTION
$IPTA -t mangle -N INPUT_SELECTION
insel="$IPTA -t mangle -A INPUT_SELECTION"

log "Connections from external interface will use same link for replies:"
for ((lnk=0; lnk < ${#LINKS[@]}; lnk++))
do
	[ "${LINKS[$lnk]}" != y ] && continue
	$insel -i ${EIF[$lnk]} -j MARK --set-mark $[$lnk+1]
	$insel -i ${EIF[$lnk]} -j ACCEPT
done

### Customize: preference for traffic
# NOTE: before routing we are not sure if packet is going to be
# forwarded or delivered locally

# CUSTOMIZE HERE

# For remote SMTP prefer LINK1 then LINK2 then some alive link:
#$insel -i INDEV -p tcp -d $MAIL --dport 25 -j LINK1
#$insel -i INDEV -p tcp -d $MAIL --dport 25 -j LINK2

# By default, routing will select link for traffic forwarded to world

## End of INPUT_SELECTION

# Connections can not change link, get link from saved mark, even for
# incoming traffic
$IPT  -t mangle -D PREROUTING -m conntrack --ctstate ESTABLISHED,RELATED -j CONNMARK --restore-mark
$IPTA -t mangle -A PREROUTING -m conntrack --ctstate ESTABLISHED,RELATED -j CONNMARK --restore-mark

# Select link if not selected yet. NEW state can match conns even after reboot.
# Even for incoming connections destined to local address.
$IPT  -t mangle -D PREROUTING -m state --state NEW -j INPUT_SELECTION
$IPTA -t mangle -A PREROUTING -m state --state NEW -j INPUT_SELECTION

## OUTPUT mangle (locally generated connection):
## OUTPUT_SELECTION table (mangle):
# - called for OUTPUT for new connections
#	- change outgoing link after routing, if needed

# Assign link after output routing:
$IPT -t mangle -F OUTPUT_SELECTION
$IPTA -t mangle -N OUTPUT_SELECTION
outsel="$IPTA -t mangle -A OUTPUT_SELECTION"

### Customize: preference for traffic

# CUSTOMIZE HERE

# For remote SMTP prefer LINK1 then LINK2 then some alive link:
#$outsel -p tcp -d $MAIL --dport 25 -j LINK1
#$outsel -p tcp -d $MAIL --dport 25 -j LINK2

## End of OUTPUT_SELECTION

$IPT  -t mangle -D OUTPUT -m state --state NEW -j OUTPUT_SELECTION
$IPTA -t mangle -A OUTPUT -m state --state NEW -j OUTPUT_SELECTION

## POSTROUTING mangle:
# - make sure connection has assigned link

log "Assign link if not done yet:"
$IPT  -t mangle -F GUESS_LINK
$IPTA -t mangle -N GUESS_LINK

# Sadly, we can not match gateway, only output device, so we can
# support only nexthops on different external interfaces
for ((lnk=0; lnk < ${#LINKS[@]}; lnk++))
do
	[ "${LINKS[$lnk]}" != y ] && continue
	$IPTA -t mangle -A GUESS_LINK -o ${EIF[$lnk]} -j MARK --set-mark $[$lnk+1]
done

log "Flush all mangle rules in POSTROUTING:"
$IPT -t mangle -F POSTROUTING

# Connection has no mark from PREROUTING, see which link is selected by routing
log "Guess link from output device:"
$IPTA -t mangle -A POSTROUTING -m state --state NEW -m mark --mark 0 -j GUESS_LINK

log "Save mark in conntrack when connection is created:"
$IPTA -t mangle -A POSTROUTING -m conntrack --ctstate NEW -j CONNMARK --save-mark

log "Flush all SNAT rules in POSTROUTING:"
$IPT -t nat -F POSTROUTING

# POSTROUTING SNAT:
log "SNAT for every link:"
for ((lnk=0; lnk < ${#LINKS[@]}; lnk++))
do
	[ "${LINKS[$lnk]}" != y ] && continue
	# If LIP!=PIP add more specific SNAT route for GW
	[ "${PIP[$lnk]}" != "${LIP[$lnk]}" ] && \
		$IPTA -t nat -A POSTROUTING -m mark --mark $[$lnk+1] \
			-o ${EIF[$lnk]} -d ${GW[$lnk]} \
			-j SNAT --to-source ${LIP[$lnk]}
	$IPTA -t nat -A POSTROUTING -m mark --mark $[$lnk+1] \
		-o ${EIF[$lnk]} -j SNAT --to-source ${PIP[$lnk]}
done

# Trigger reset of rules in LINK<n> tables
rm -f "$s"

exit_func
trap "" EXIT

log "Checking links..."

# Do health checks, they will add rules in LINKn> and route tables
$TEST $dp/mpath_connmark_check.sh "$cfg" fast

exit 0

