#!/bin/bash
# Recreate the default route and the iptables rules depending on
# alive links in priority groups
# Arg 1: new bitmask of alive links
# Arg 2: old bitmask of alive links
# Arg 3: config file
#
# Called by mpath_connmark_check.sh (as $c) after current check of links
if [ "$1" = "" -o "$2" = "" ]
then
	echo "Missing Parameters!"
	exit 1
fi

# m: mask of alive links after current check
m="$1"
links="0"

IPT=iptables

cfg="$3"
if ! [ -e "$cfg" ]
then
	echo "Missing configuration file: $cfg"
	exit 1
fi

. "$cfg"

lm="0"
# For all priority groups
for ((prio=0; prio<${#PRIO_LINKS[*]}; prio++))
do
	# Group contains alive link?
	if [ "$[${PRIO_LINKS[$prio]} & $m]" -ne 0 ]
	then
		# Stop on the first group with alive links
		lm="${PRIO_LINKS[$prio]}"
		break
	fi
done

# Disable links not allowed for priority
# ma: mask of alive links after current check, restricted to contain only
# links from the first group with alive link
ma="$[$m & $lm]"

exit_func() {
	rm -f "$LOCK"
}

# Back in time?
find $LOCK -mmin -0 -exec rm {} \; &> /dev/null
# Lock rules
lockfile -l 120 -s 1 -r 0 -1 $LOCK &> /dev/null || exit 1

trap exit_func EXIT

NH=""
for ((lnk=0; lnk < ${#LINKS[@]}; lnk++))
do
	[ "${LINKS[$lnk]}" != y ] && continue
	# Make sure specific route is present for registered devices,
	# even for dead links. This is needed to allow ping checks
	# to work.
	# Routes can disappear if devices are unregistered, so we have to
	# add them on every check. "add" is enough for this but we want
	# any config changes to be applied, so use "replace".
	ip route replace default via ${GW[$lnk]} dev ${EIF[$lnk]} \
		src ${PIP[$lnk]} table $[$lnk+1]
	ip route add prohibit default metric 1 table $[$lnk+1] 2> /dev/null
	# Alive link from the alive priority group?
	if [ "$[$ma & (1 << $lnk)]" -ne 0 ]
	then
		w1="${WEIGHT[$lnk]}"
		[ "$w1" = "" ] && w1="1"
		NH="$NH nexthop via ${GW[$lnk]} dev ${EIF[$lnk]} weight $w1"
	fi
done
# Apply all alive links from the alive priority group to table 100
[ "$NH" != "" ] && ip route replace default table 100 $NH
[ "$NH" = "" ] && ip route del default table 100
ip route add prohibit default metric 1 table 100 2> /dev/null

# For iptables we will ignore the priority groups, we will keep
# actual rules in tables LINK<n>, only for alive links.

reset=""
o="$2"
[ "$o" -lt 0 ] && o=0 && reset=y

# Alive links changed?
diff="$[$m ^ $o]"
for ((lnk=0; lnk < ${#LINKS[@]}; lnk++))
do
	lid="$[$lnk+1]"
	[ "$reset" != "" ] && $IPT -t mangle -F "LINK$lid"
	if [ "$[$diff & (1 << $lnk)]" -ne 0 ]
	then
		if [ "$[$m & (1 << $lnk)]" -ne 0 ]
		then
			$IPT -t mangle -A "LINK$lid" -j MARK --set-mark "$lid"
			$IPT -t mangle -A "LINK$lid" -j ACCEPT
		else
			[ "$reset" = "" ] && $IPT -t mangle -F "LINK$lid"
		fi
	fi
done

echo "$m"

exit 0

